# returns (2c-1)*(2c-3)*...*3*1
def odd_factorial(c, to_mult = 1):
    if c<2:
        return to_mult
    return odd_factorial(c-1, to_mult*(2*c-1))

def enum_Levis_BC(n):
    res = []
    for k in range(n+1):
        for c in range(int((n-k)/2)+1):
            res.append([n-k-2*c,c,k])
    return res

# up to outer conjugation
# for SO_4n, i.e. D_2n
def enum_Levis_D(n):
    res = []
    for k in range(n+1):
        for c in range(n-k+1):
            res.append([n-k-c,c,k])
    return res

# c >= 0
# 0 <= x < (2c-1)*(2c-3)*...*3*1
def disjoint_pairs_from_nb(c, x, init_set = None, begin_res = []):
    if c==0:
        return begin_res
    if not init_set:
        init_set = list(range(1, 2*c+1))
    second_el_idx, new_x = divmod(x, odd_factorial(c-1))
    first_pair = [init_set.pop(0)]
    first_pair.append(init_set.pop(second_el_idx))
    return disjoint_pairs_from_nb(c-1, new_x, init_set, begin_res + first_pair)

def enum_sigma_repr_BC(Levi):
    res = []
    n = Levi[0]+2*Levi[1]+Levi[2]
    for sigma_a, sigma_cd, sign0 in set_partitions_2(n, Levi[0]):
        for sigma_c, sigma_d, sign1 in set_partitions_2(n-Levi[0], 2*Levi[1],
                                                        sigma_cd):
            res.append([sigma_a, sigma_c, sigma_d, sign0+sign1])
    return res

def enum_sigma_repr_D(Levi):
    res = []
    n = Levi[0]+Levi[1]+Levi[2]
    for sigma_a, sigma_cd, sign0 in set_partitions_2(2*n, 2*Levi[0]):
        for sigma_c, sigma_d, sign1 in set_partitions_2(2*(n-Levi[0]),
                                                        2*Levi[1], sigma_cd):
            res.append([sigma_a, sigma_c, sigma_d, sign0+sign1])
    return res


# computes the contribution of the elements  +- 1 in the GL_1^t part of the Levi
#   includes the contribution to (-1/2)^(dim A_M)*|W_M|/|W_G|*|Levi's|
# returns a list L such that if x is the number of i in [1,n] for which
# k_s(i)+s(i)+i is odd, L[x] is the contribution
def contrib_plusminusone(t):
    res_list = []
    for x in range(t+1):
        res = 0
        pm_list = ([-1]*x)+([1]*(t-x))
        for b in range(int(t/2)+1):
            sub_res = 0
            for parti in set_partitions_2(t, 2*b):
                sub_res += prod([pm_list[i-1] for i in parti[0]])
            res += (-1)^b*sub_res
        res_list.append(res*(-1)^(t*(t-1)/2)/2^t)
    return res_list

pre_contr_pm1 = [contrib_plusminusone(t) for t in range(11)]

def contrib_Levi_SO_odd(L, ell_tr, wt_bounds, extra_wts):
    a = L[0]
    c = L[1]
    d = L[2]
    n = a+2*c+d
    sigma_repr_list = enum_sigma_repr_BC(L)
    nb_parti_pairs = odd_factorial(c)
    parti_pairs = [disjoint_pairs_from_nb(c, x) for x in range(nb_parti_pairs)]
    res = {}
    dom_weight = [0]*n
    for dom_weight in chain_gen(it_dw_SO_odd(n, *wt_bounds), extra_wts):
        weight_res = 0
        for sigma in sigma_repr_list:
            num_minusone = 0
            for i in range(a):
                if Mod(dom_weight[sigma[0][i]-1]+sigma[0][i]+i,2)==Mod(0,2):
                    num_minusone += 1
            GL2_contrib = 0
            for x in range(nb_parti_pairs):
                GL2_pairs_contrib = (-1)^x
                for i in range(c):
                    s1 = sigma[1][parti_pairs[x][2*i]-1]
                    s2 = sigma[1][parti_pairs[x][2*i+1]-1]
                    k1 = dom_weight[s1-1]-dom_weight[s2-1]-s1+s2-1
                    k2 = dom_weight[s1-1]+dom_weight[s2-1]+2*n-s1-s2
                    if Mod(k1,2)==Mod(0,2):
                        GL2_pairs_contrib *= ell_tr[1][(k1/2,)]
                    else:
                        GL2_pairs_contrib *= -ell_tr[1][(k2/2,)]
                GL2_contrib += GL2_pairs_contrib
            SO_2dplus1_weight = [dom_weight[sigma[2][i]-1]+n-d+i+1-sigma[2][i]
                                 for i in range(d)]
            weight_res += (-1)^sigma[3]*GL2_contrib\
                *pre_contr_pm1[a][num_minusone]\
                *ell_tr[d][tuple(SO_2dplus1_weight)]
        res[dom_weight] = weight_res
    return res

def contrib_Levi_Sp(L, ell_tr, ell_tr_PGL2, wt_bounds, extra_wts):
    a = L[0]
    c = L[1]
    d = L[2]
    n = a+2*c+d
    sigma_repr_list = enum_sigma_repr_BC(L)
    nb_parti_pairs = odd_factorial(c)
    parti_pairs = [disjoint_pairs_from_nb(c, x) for x in range(nb_parti_pairs)]
    res = {}
    for dom_weight in chain_gen(it_dw_Sp(n, *wt_bounds), extra_wts):
        weight_res = 0
        for sigma in sigma_repr_list:
            num_minusone = 0
            for i in range(a):
                if Mod(dom_weight[sigma[0][i]-1]+sigma[0][i]+i,2)==Mod(0,2):
                    num_minusone += 1
            if (Mod(a,2)==Mod(0,2) and num_minusone==a)\
                    or (Mod(a,2)==Mod(1,2) and num_minusone==0):
                GL2_contrib = 0
                for x in range(nb_parti_pairs):
                    GL2_pairs_contrib = (-1)^x
                    for i in range(c):
                        s1 = sigma[1][parti_pairs[x][2*i]-1]
                        s2 = sigma[1][parti_pairs[x][2*i+1]-1]
                        k1 = dom_weight[s1-1]-dom_weight[s2-1]-s1+s2-1
                        k2 = dom_weight[s1-1]+dom_weight[s2-1]+2*n+1-s1-s2
                        if Mod(k1, 2)==Mod(0, 2):
                            GL2_pairs_contrib *= (ell_tr_PGL2[(k1/2,)]
                                                  -ell_tr_PGL2[(k2/2,)])
                        else:
                            GL2_pairs_contrib = 0
                            break
                    GL2_contrib += GL2_pairs_contrib
                Sp_2d_weight = [dom_weight[sigma[2][i]-1]+n-d+i+1-sigma[2][i]
                                for i in range(d)]
                if Mod(sum(Sp_2d_weight), 2)==Mod(0, 2):
                    weight_res += (-1)^sigma[3]*(-1)^(a*(a-1)/2)*GL2_contrib\
                        *ell_tr[d][tuple(Sp_2d_weight)]
        res[dom_weight] = weight_res
    return res

def contrib_Levi_SO_even(L, ell_tr, ell_tr_PGL2, wt_bounds, extra_wts):
    a = L[0]
    c = L[1]
    d = L[2]
    n = a+c+d # beware: half the rank!
    sigma_repr_list = enum_sigma_repr_D(L)
    nb_parti_pairs = odd_factorial(c)
    parti_pairs = [disjoint_pairs_from_nb(c, x) for x in range(nb_parti_pairs)]
    res = {}
    for dom_weight in chain_gen(it_dw_SO_even(2*n, *wt_bounds), extra_wts):
        weight_res = 0
        for sigma in sigma_repr_list:
            num_minusone = 0
            for i in range(2*a):
                if Mod(dom_weight[sigma[0][i]-1]+sigma[0][i]+i,2)==Mod(0,2):
                    num_minusone += 1
            if num_minusone==2*a or num_minusone==0:
                GL2_contrib = 0
                for x in range(nb_parti_pairs):
                    GL2_pairs_contrib = (-1)^x
                    for i in range(c):
                        s1 = sigma[1][parti_pairs[x][2*i]-1]
                        s2 = sigma[1][parti_pairs[x][2*i+1]-1]
                        k1 = dom_weight[s1-1]-dom_weight[s2-1]-s1+s2-1
                        k2 = dom_weight[s1-1]+dom_weight[s2-1]+4*n-1-s1-s2
                        if Mod(k1, 2)==Mod(0, 2):
                            GL2_pairs_contrib *= (ell_tr_PGL2[(k1/2,)]
                                                  +ell_tr_PGL2[(k2/2,)])
                        else:
                            GL2_pairs_contrib = 0
                            break
                    GL2_contrib += GL2_pairs_contrib
                SO_4d_weight = [dom_weight[sigma[2][i]-1]
                                +2*n-2*d+i+1-sigma[2][i] for i in range(2*d)]
                if Mod(sum(SO_4d_weight), 2)==Mod(0, 2):
                    weight_res += (-1)^sigma[3]*GL2_contrib\
                        *ell_tr[d][tuple(SO_4d_weight)]
        if a>0:
            weight_res /= 2
        res[dom_weight] = weight_res
    return res

def EPchar_SO_odd(n, ell_tr, wt_bounds, extra_wts, minus_elliptic=False):
    res = {dw : 0 for dw in chain_gen(it_dw_SO_odd(n, *wt_bounds), extra_wts) }
    for L in enum_Levis_BC(n):
        if minus_elliptic and L[2]==n:
            continue
        L_contrib = contrib_Levi_SO_odd(L, ell_tr, wt_bounds, extra_wts)
        for dw, cont in L_contrib.items():
            res[dw] += cont
    return res

def EPchar_Sp(n, ell_tr, ell_tr_PGL2, wt_bounds, extra_wts,
              minus_elliptic=False):
    res = {dw : 0 for dw in chain_gen(it_dw_Sp(n, *wt_bounds), extra_wts) }
    for L in enum_Levis_BC(n):
        if minus_elliptic and L[2]==n:
            continue
        L_contrib = contrib_Levi_Sp(L, ell_tr, ell_tr_PGL2, wt_bounds,
                                    extra_wts)
        for dw, cont in L_contrib.items():
            res[dw] += cont
    return res

def EPchar_SO_even(n, ell_tr, ell_tr_PGL2, wt_bounds, extra_wts,
                   minus_elliptic=False):
    res = {dw : 0 for dw in chain_gen(it_dw_SO_even(n, *wt_bounds), extra_wts) }
    for L in enum_Levis_D(Integer(n/2)):
        if minus_elliptic and L[2]==n/2:
            continue
        L_contrib = contrib_Levi_SO_even(L, ell_tr, ell_tr_PGL2, wt_bounds,
                                         extra_wts)
        for dw, cont in L_contrib.items():
            res[dw] += cont
    return res
