# 1 -> 2nd elt, 2 -> 3rd, 4 -> 4th, etc
# ss is the mask for the second set
# the first set always contains the first element
# I is assumed non-empty
def special_subset_binary(I, ss):
    first_set = [I[0]]
    second_set = []
    mask = 1
    for x in I[1:]:
        if mask&ss:
            second_set.append(x)
        else:
            first_set.append(x)
        mask <<= 1
    return first_set, second_set

# enumerate the partitions of a set into non-empty subsets
# partitions are not ordered
def set_partitions(init_set):
    n = len(init_set)
    if n==0:
        yield []
        return
    for ss in range(1<<(n-1)):
        first_set, second_set = special_subset_binary(init_set, ss)
        for A in set_partitions(second_set):
            yield [first_set]+A
    return

def multi_set_partitions(set_list):
    if len(set_list)==0:
        yield []
        return
    first_set = set_list[0]
    for x in multi_set_partitions(set_list[1:]):
        for first_parti in set_partitions(first_set):
            yield [first_parti]+x
    return

# same but up to max_odd subsets are set aside, and only these may have odd
# cardinality
def set_partitions_parity(init_set, max_odd):
    n = len(init_set)
    if n==0:
        yield [], [[] for _ in range(max_odd)]
        return
    for ss in range(1<<(n-1)):
        first_set, second_set = special_subset_binary(init_set, ss)
        if max_odd<=0 and len(first_set)%2:
            continue
        if not(len(first_set)%2):
            for A, B in set_partitions_parity(second_set, max_odd):
                yield [first_set]+A, B
        if max_odd>0:
            for A, B in set_partitions_parity(second_set, max_odd-1):
                yield A, [first_set]+B
    return

# enumerate products of partitions of sets (d_indices is a list of set,
# corresponding to a list of distinct integers d in d_list), some of which
# (corresponding to odd d) are split in two as in set_partitions_parity (max_odd
# is the number of times that d occurs in spe_d)
def multi_set_partitions_parity(d_list, d_indices, spe_d, parity, i=0):
    if len(d_list)==0:
        yield [], [[] for _ in spe_d]
        return
    d = d_list[0]
    ind = d_indices[0]
    if d%2==parity:
        spe_d_ind = [j for j, x in enumerate(spe_d) if x==d]
        for A, B in multi_set_partitions_parity(d_list[1:], d_indices[1:],
                                                spe_d, parity, i+1):
            for first, spe_part in set_partitions_parity(ind, len(spe_d_ind)):
                for k, j in enumerate(spe_d_ind):
                    B[j] = spe_part[k]
                yield [first]+A, B
    else:
        for A, B in multi_set_partitions_parity(d_list[1:], d_indices[1:],
                                                spe_d, parity, i+1):
            for first in set_partitions(ind):
                yield [first]+A, B
    return

def char_Lpacket_SO_odd(n):
    res = []
    a = int((n+1)/2)
    b = int(n/2)
    for i in range(b+1):
        for set1 in [x[0] for x in set_partitions_2(a, i, list(range(a)))]:
            for set2 in [y[0] for y in set_partitions_2(b, i, list(range(b)))]:
                A = [Mod(0, 2)]*n
                for z in set1:
                    A[2*z] = Mod(1, 2)
                for z in set2:
                    A[2*z+1] = Mod(1, 2)
                res.append(A)
    return res

def char_Lpacket_SO_even(n):
    res = []
    a = Integer(n/2)
    for i in range(a+1):
        for set1 in [x[0] for x in set_partitions_2(a, i, list(range(a)))]:
            for set2 in [y[0] for y in set_partitions_2(a, i, list(range(a)))]:
                A = [Mod(0, 2)]*n
                for z in set1:
                    A[2*z] = Mod(1, 2)
                for z in set2:
                    A[2*z+1] = Mod(1, 2)
                res.append(A)
    return res

def char_Lpacket_Sp(n):
    if n==0:
        return [[]]
    res = []
    for x in char_Lpacket_Sp(n-1):
        res.append([Mod(0,2)]+x)
        res.append([Mod(1,2)]+x)
    return res

def formal_archi_Arthur_param(n, b=-1):
    if b<0 or (n>0 and n<=b+1):
        yield [[], n]
    for d in range(1, n+1):
        for x in formal_archi_Arthur_param(n-d, b):
            yield [[d]+x[0], x[1]]
    return

def mask_of_formal_archi_Arthur_param(P):
    res = 0
    cur_m = 1
    for i in range(P[1]):
        res |= cur_m
        cur_m *= 2
    for d in P[0][::-1]:
        cur_m *= 2
        for i in range(d-1):
            res |= cur_m
            cur_m *= 2
    return res

# Input: formal archimedean Arthur parameter P, discrete series given as its
# tempered character temp_char
# Output: sign in Johnson's resolution and character on non-tempered S_Psi
def char_and_sign_in_formal_archi_Arthur_param_SO_odd(P, temp_char):
    cur_pos = 0
    non_temp_char = []
    sign_Johnson = Mod(0, 2)
    for d in P[0]:
        a = len([i for i in range(d)
                 if Mod(temp_char[cur_pos+i]+i, 2)==Mod(0, 2)])
        sign_Johnson += a*(d-a)
        non_temp_char.append(sum(temp_char[cur_pos:cur_pos+d]))
        cur_pos += d
    d = P[1]
    a = 2*len([i for i in range(d)
               if Mod(temp_char[cur_pos+i]+i, 2)==Mod(0, 2)])
    if Mod(d, 2)==Mod(0, 2):
        a += 1
    sign_Johnson += Mod((a*(2*d+1-a))/2, 2)
    if d>0:
        non_temp_char.append(sum(temp_char[cur_pos:cur_pos+d]))
    return tuple(non_temp_char), sign_Johnson

def char_and_sign_in_formal_archi_Arthur_param_SO_even(P, temp_char):
    cur_pos = 0
    non_temp_char = []
    sign_Johnson = Mod(0, 2)
    for d in P[0]:
        a = len([i for i in range(d)
                 if Mod(temp_char[cur_pos+i]+i, 2)==Mod(0, 2)])
        sign_Johnson += a*(d-a)
        non_temp_char.append(sum(temp_char[cur_pos:cur_pos+d]))
        cur_pos += d
    d = P[1]
    if d>0:
        non_temp_char.append(sum(temp_char[cur_pos:cur_pos+d]))
    return tuple(non_temp_char), sign_Johnson

def char_and_sign_in_formal_archi_Arthur_param_Sp(P, temp_char):
    cur_pos = 0
    non_temp_char = []
    sign_Johnson = Mod(P[1]*(P[1]+1)/2, 2)
    for d in P[0]:
        a = len([i for i in range(d)
                 if Mod(temp_char[cur_pos+i]+i, 2)==Mod(0, 2)])
        sign_Johnson += a*(d-a)
        non_temp_char.append(sum(temp_char[cur_pos:cur_pos+d]))
        cur_pos += d
    return tuple(non_temp_char), sign_Johnson

def formal_archi_Arthur_param_with_char_SO_odd(n):
    discrete_series = char_Lpacket_SO_odd(n)
    for P in formal_archi_Arthur_param(n):
        char_with_mult = {}
        for ds in discrete_series:
            non_temp_char, sign_Johnson =\
                char_and_sign_in_formal_archi_Arthur_param_SO_odd(P, ds)
            if not (non_temp_char in char_with_mult):
                char_with_mult[non_temp_char] = (-1)^sign_Johnson
            else:
                char_with_mult[non_temp_char] += (-1)^sign_Johnson
        yield [P, char_with_mult]
    return

def formal_archi_Arthur_param_with_char_SO_odd_cpct(n):
    discrete_series = [[Mod(i+n, 2) for i in range(n)]]
    for P in formal_archi_Arthur_param(n):
        char_with_mult = {}
        for ds in discrete_series:
            non_temp_char, sign_Johnson =\
                char_and_sign_in_formal_archi_Arthur_param_SO_odd(P, ds)
            if not (non_temp_char in char_with_mult):
                char_with_mult[non_temp_char] = (-1)^sign_Johnson
            else:
                char_with_mult[non_temp_char] += (-1)^sign_Johnson
        yield [P, char_with_mult]
    return

def formal_archi_Arthur_param_with_char_SO_even(n, split=True):
    if split:
        discrete_series = char_Lpacket_SO_even(n)
    else: # signature (2n-2,2) (n odd) or (2n,0) (n even)
        discrete_series = [[Mod(i,2) for i in range(n)]]
    for P in formal_archi_Arthur_param(n):
        char_with_mult = {}
        for ds in discrete_series:
            non_temp_char, sign_Johnson =\
                char_and_sign_in_formal_archi_Arthur_param_SO_even(P, ds)
            if not (non_temp_char in char_with_mult):
                char_with_mult[non_temp_char] = (-1)^sign_Johnson
            else:
                char_with_mult[non_temp_char] += (-1)^sign_Johnson
        yield [P, char_with_mult]
    return

def formal_archi_Arthur_param_with_char_Sp(n):
    discrete_series = char_Lpacket_Sp(n)
    for P in formal_archi_Arthur_param(n):
        char_with_mult = {}
        for ds in discrete_series:
            non_temp_char, sign_Johnson =\
                char_and_sign_in_formal_archi_Arthur_param_Sp(P, ds)
            if not (non_temp_char in char_with_mult):
                char_with_mult[non_temp_char] = (-1)^sign_Johnson
            else:
                char_with_mult[non_temp_char] += (-1)^sign_Johnson
        yield [P, char_with_mult]
    return

def formal_archi_Arthur_param_with_char_Siegel(n):
    # char of centralizer for holomorphic discrete series
    ds = [Mod(i,2) for i in range(n)]
    for P in formal_archi_Arthur_param(n):
        if P[1]==0:
            # holomorphic discrete series is part of the AJ packet
            non_temp_char, _ =\
                char_and_sign_in_formal_archi_Arthur_param_Sp(P, ds)
            yield [P, {non_temp_char:1}]
    return

def formal_archi_Arthur_param_with_char_Siegel_low(n, max_k):
    # char of centralizer for holomorphic discrete series
    ds = [Mod(i,2) for i in range(n)]
    for P in formal_archi_Arthur_param(n):
        if P[1]>=n-max_k or (len(P[0])>0 and P[0][-1]>=2*(n-max_k)+1):
            # when the weight is small wrt n we need big [d]'s
            non_temp_char, _ =\
                char_and_sign_in_formal_archi_Arthur_param_Sp(P, ds)
            yield [P, {non_temp_char:1}]
    return

# for low (<=genus) scalar weight Siegel modular forms we also need to consider
# parameters having 3 odd factors
def formal_archi_Arthur_param_with_char_Siegel_low_3odd(n, max_k):
    # char of centralizer for holomorphic discrete series
    ds = [Mod(i,2) for i in range(n)]
    # 2n+1 = 2*?+1+2a-1+2b+1 = 2*(?+a+b)+1 and a>0, ?>=0
    # n-k = b (case H1) or b-1 (case H2)
    for b in range(max(0, n-max_k), n):
        # enumerate P = [list of d_i's, a] with a<=b+1
        for P in formal_archi_Arthur_param(n-b, b):
            non_temp_char, _ =\
                char_and_sign_in_formal_archi_Arthur_param_SO_even(P, ds)
            yield [[P[0], [P[1], b]], non_temp_char]
    return

# enumerate (yield) the possible formal global Arthur parameters for SO_{2n+1}
# or Sp_2n which are Adams-Johnson at the real place.
# "formal" means that we consider parameters sum_i pi_i[d_i] without specifying
# the precise Hodge weights of the pi_{i,infty}, only how they are ordered (the
# w_l's below are decreasing and positive).
# output format: yields [A, formal Archimedean parameter, dict of EP char
# according to global character of S_psi], with A as follows:
# * a list of lists:
#   * d_i
#   * list of indices l such that I_{w_l} appears in the Langlands parameter of
#     pi_{i,infty} where pi_i[d_i] is an even factor
# * an integer a
# * B
# for Sp_2n, i.e. odd orthogonal parameters, B is a list of indices l such that
# I_{w_l} appears in the Langlands parameter of pi_{i0, infty}, where
# pi_i0[2a+1] is the unique odd factor in the parameter;
# for SO_{2n+1}, i.e. symplectic parameters, B is a list which is empty if a=0,
# and a list of indices l such that I_{w_l} appears in the Langlands parameter
# of pi_{i0, infty}, where pi_i0[2a] is the (unique if it exists) factor with
# pi_i0 odd orthogonal.
def formal_global_Arthur_param_with_char_BC(n, group_type, max_k=100):
    if group_type=='B': # SO_odd
        it_archi_param = formal_archi_Arthur_param_with_char_SO_odd(n)
        parity = 0
    elif group_type=='B_cpct':
        it_archi_param = formal_archi_Arthur_param_with_char_SO_odd_cpct(n)
        parity = 0
    elif group_type=='C': # Sp
        it_archi_param = formal_archi_Arthur_param_with_char_Sp(n)
        parity = 1
    elif group_type=='Siegel':
        it_archi_param = formal_archi_Arthur_param_with_char_Siegel(n)
        parity = 1
    elif group_type=='Siegel_low':
        it_archi_param = formal_archi_Arthur_param_with_char_Siegel_low(n,
                                                                        max_k)
        parity = 1
    else:
        print("Error in formal_global_Arthur_param_with_char_BC: group_type must be B, C, or Siegel.")
        return
    for archi_param, archi_char_dict in it_archi_param:
        positions = [0]
        cur_pos = 0
        for d in archi_param[0]:
            cur_pos += d
            positions.append(cur_pos)
        d_list = list(set(archi_param[0]))
        num_d = len(d_list)
        d_indices = [[i for i in range(len(archi_param[0]))
                      if archi_param[0][i]==d] for d in d_list]
        if 2*archi_param[1]+parity>0:
            spe_d = (2*archi_param[1]+parity,)
        else:
            spe_d = ()
        if not all(len(d_indices[i])%2==0 for i, d in enumerate(d_list)
                                          if d%2==parity and not(d in spe_d)):
            continue
        if d_list==[1] and all(d==1 for d in spe_d):
            tempered = True
        else:
            tempered = False
        for main_parti, spe_parti in\
                multi_set_partitions_parity(d_list, d_indices, spe_d, parity):
            if parity==0 and tempered and len(main_parti[0])==1:
                # tempered non-endoscopic symplectic parameter
                continue
            if parity==1 and tempered and len(main_parti[0])==0:
                # tempered non-endoscopic odd orthogonal parameter
                continue
            global_char_dict = {}
            for archi_char in archi_char_dict:
                global_char_as_list = []
                for i in range(num_d):
                    for L in main_parti[i]:
                        global_char_as_list.append(sum([archi_char[j]
                                                        for j in L]))
                # symplectic parameter containing char[even]
                if parity==0 and archi_param[1]>0:
                    global_char_as_list.append(sum([archi_char[j]
                                                    for j in spe_parti[0]])
                                               +archi_char[len(archi_param[0])])
                global_char = tuple(global_char_as_list)
                if global_char in global_char_dict:
                    global_char_dict[global_char] += archi_char_dict[archi_char]
                else:
                    global_char_dict[global_char] = archi_char_dict[archi_char]
            param_main = []
            for i in range(num_d):
                for L in main_parti[i]:
                    param_main.append([d_list[i], [positions[j] for j in L]])
            if len(spe_d)==1:
                param_spe = [positions[j] for j in spe_parti[0]]
            else:
                param_spe = [[positions[j] for j in L] for L in spe_parti]
            yield [[param_main, archi_param[1], param_spe], archi_param,
                   global_char_dict]
    return

# TODO: merge with cases B and C above now that they have been simplified
# similar to the above, except that B is:
# * absent if a=0 (no odd factor in the parameter)
# * [B1, B2] if a>0, where Bk is a list of indices l such that I_{w_l} appears
# in the Langlands parameter of pi_{ik, infty} (resp), where pi_i1[1] and
# pi_i2[2a-1] are the two odd factors in the parameter.  If a=1, the two are
# distinguished by imposing dim pi_i2 = 1 mod 4 (i.e. len(B2) is even).
def formal_global_Arthur_param_with_char_SO_even(n, split=True):
    it_archi_param = formal_archi_Arthur_param_with_char_SO_even(n, split)
    for archi_param, archi_char_dict in it_archi_param:
        positions = [0]
        cur_pos = 0
        for d in archi_param[0]:
            cur_pos += d
            positions.append(cur_pos)
        d_list = list(set(archi_param[0]))
        num_d = len(d_list)
        d_indices = [[i for i in range(len(archi_param[0]))
                      if archi_param[0][i]==d] for d in d_list]
        if archi_param[1]==0:
            # no odd factor in parameter
            spe_d = ()
        else:
            # two odd factors (pi_1[1] and pi_2[d] with d=2*archi_param[1]-1)
            spe_d = (1, 2*archi_param[1]-1)
        if not all(Mod(len(d_indices[i]),2)==0
                   for i, d in enumerate(d_list) if d%2 and not(d in spe_d)):
            # necessary condition for existence of at least one partition
            # satisfying parity conditions
            continue
        for main_parti, spe_parti in\
                multi_set_partitions_parity(d_list, d_indices, spe_d, 1):
            global_char_dict = {}
            for archi_char in archi_char_dict:
                global_char_as_list = []
                for i in range(num_d):
                    for L in main_parti[i]:
                        global_char_as_list.append(sum([archi_char[j]
                                                        for j in L]))
                #if archi_param[1]>0:
                #    global_char_as_list.append(
                #        archi_char[len(archi_param[0])])
                global_char = tuple(global_char_as_list)
                if global_char in global_char_dict:
                    global_char_dict[global_char] +=\
                        archi_char_dict[archi_char]
                else:
                    global_char_dict[global_char] =\
                        archi_char_dict[archi_char]
            param_main = []
            for i in range(num_d):
                d = d_list[i]
                for L in main_parti[i]:
                    param_main.append([d, [positions[j] for j in L]])
            if archi_param[1]==0:
                if d_list==[1] and len(main_parti[0])==1: 
                    # we don't want the tempered non-endoscopic parameter
                    continue
                yield [[param_main, 0], archi_param, global_char_dict]
                continue
            param_spe = [[positions[j] for j in L] for L in spe_parti]
            yield [[param_main, archi_param[1], param_spe], archi_param,
                    global_char_dict]
    return

# enumerate (yield) the possible formal global Arthur parameters for Siegel
# square-integrable forms of scalar weight <=min(n, max_k) having three odd
# factors
# output format: yields [A, formal Archimedean parameter, char of S_psi]
# with A as follows:
# * a list of lists:
#   * d_i
#   * list of indices l such that I_{w_l} appears in the Langlands parameter of
#     pi_{i,infty} where pi_i[d_i] is an even factor
# * a list [a, b] with 0<=a-1<=b
# * a list [B1, B2, B3]
# * True or False according to whether we are in case H1 or H2
# Bi is a list of indices l such that I_{w_l} appears in the Langlands parameter
# of pi_{i, infty}, where pi_1[1], pi_2[2a+1] and pi_3[2b+1] are the odd factors
# in the parameter.
def formal_global_Arthur_param_with_char_Siegel_low_3odd(n, max_k=100):
    it_archi_param =\
            formal_archi_Arthur_param_with_char_Siegel_low_3odd(n, max_k)
    for archi_param, archi_char in it_archi_param:
        positions = [0]
        cur_pos = 0
        for d in archi_param[0]:
            cur_pos += d
            positions.append(cur_pos)
        d_list = list(set(archi_param[0]))
        num_d = len(d_list)
        d_indices = [[i for i in range(len(archi_param[0]))
                      if archi_param[0][i]==d] for d in d_list]
        a, b = archi_param[1]
        # odd factors ?, ?[2a-1] and ?[2b+1] in the Archimedean parameter, where
        # ? are characters
        spe_d = (1, 2*a-1, 2*b+1)
        if not all(Mod(len(d_indices[i]),2)==0
                   for i, d in enumerate(d_list) if d%2 and not(d in spe_d)):
            continue
        for main_parti, spe_parti in\
                multi_set_partitions_parity(d_list, d_indices, spe_d, 1):
            # n = b+k' and epsilon^{k'}[2(n-k')+1] has to appear in the
            # parameter.
            # this could be caught sooner, at the cost of making
            # multi_set_partitions_parity more complicated
            if (n+b+len(spe_parti[2]))%2:
                if a==b+1 and (len(spe_parti[1])+len(spe_parti[2]))%2:
                    spe_parti = [spe_parti[0], spe_parti[2], spe_parti[1]]
                elif b==0 and (len(spe_parti[0])+len(spe_parti[2]))%2:
                    spe_parti = [spe_parti[2], spe_parti[1], spe_parti[0]]
                else:
                    continue
            if a==1 and len(spe_parti[0])+len(spe_parti[1])==0:
                # parameter contains [1]+[1] and so is not discrete
                continue
            if a==b+1 and len(spe_parti[2])+len(spe_parti[1])==0:
                # parameter contains [2b+1]+[2b+1] and so is not discrete
                continue
            global_char = []
            for i in range(num_d):
                for L in main_parti[i]:
                    global_char.append(sum([archi_char[j] for j in L]))
            # the following is the value at s_{i_1 i_2}
            global_char.append(sum([archi_char[j] for j in spe_parti[0]])
                               +sum([archi_char[j] for j in spe_parti[1]])
                               +archi_char[len(archi_param[0])])
            param_even = []
            for i in range(num_d):
                d = d_list[i]
                for L in main_parti[i]:
                    param_even.append([d, [positions[j] for j in L]])
            param_odd = [[positions[j] for j in L] for L in spe_parti]
            char_extra = Mod(sum(archi_char[j] for j in spe_parti[1])
                             +sum(archi_char[j] for j in spe_parti[2]),2)
            if (len(spe_parti[1])+len(spe_parti[2]))%2:
                # condition P_{i0 i2} holds
                char_extra += n-b+1
            yield [[param_even, archi_param[1], param_odd, True], archi_param,
                   global_char+[char_extra]]
            # check for possible "hole" in weights in case H2
            if n-2*b in positions and n-b+1<=max_k:
                yield [[param_even, archi_param[1], param_odd, False],
                       archi_param, global_char+[1+char_extra]]
    return

def epsilon_archi_product(H1, H2):
    res = Mod(0, 2)
    for w1 in H1:
        for w2 in H2:
            res += Mod(1+2*max(w1, w2),2)
    return res

# computes the endoscopic contribution, for a given weight and given formal
# global parameter
def endo_contrib_SO_odd(n, dom_weight, para, char_dict,
                        new_sympl_para_list, new_orth_odd_para_list,
                        new_orth_even_para_list):
    sub_Hodge_weights = []
    epsilon_Psi_as_list = []
    for d, pos_list in para[0]:
        sub_H_wt = [dom_weight[pos]+n-pos-d/2 for pos in pos_list]
        sub_Hodge_weights.append(sub_H_wt)
    if para[1]==0: # no (odd orth)[even] in the parameter
        for x in range(len(para[0])):
            d = para[0][x][0]
            if Mod(d,2)==Mod(1,2):
                epsilon_Psi_as_list.append(Mod(sum(
                    epsilon_archi_product(sub_Hodge_weights[x],
                                          sub_Hodge_weights[y])
                    for y in range(len(para[0]))
                    if Mod(para[0][y][0],2)==Mod(0, 2) and para[0][y][0]>d), 2))
            else:
                epsilon_Psi_as_list.append(Mod(sum(
                    epsilon_archi_product(sub_Hodge_weights[x],
                                          sub_Hodge_weights[y])
                    for y in range(len(para[0]))
                    if Mod(para[0][y][0],2)==Mod(1, 2) and para[0][y][0]<d), 2))
    else: # one component of type (odd orth)[even] in the parameter
        orth_odd_Hodge_weights = [dom_weight[pos]+n-pos-para[1]
                                  for pos in para[2]]
        for x in range(len(para[0])):
            d = para[0][x][0]
            if Mod(d,2)==Mod(1,2):
                if d<2*para[1]:
                    epsilon_Psi_as_list.append(
                        epsilon_archi_product(sub_Hodge_weights[x],
                                              orth_odd_Hodge_weights)\
                        +Mod(sum(w+1/2 for w in sub_Hodge_weights[x]), 2)\
                        +Mod(sum(epsilon_archi_product(sub_Hodge_weights[x],
                                                       sub_Hodge_weights[y])
                                 for y in range(len(para[0]))
                                 if Mod(para[0][y][0],2)==Mod(0, 2)
                                 and para[0][y][0]>d), 2))
                else:
                    epsilon_Psi_as_list.append(Mod(sum(
                        epsilon_archi_product(sub_Hodge_weights[x],
                                              sub_Hodge_weights[y])
                        for y in range(len(para[0]))
                        if Mod(para[0][y][0],2)==Mod(0, 2)
                        and para[0][y][0]>d), 2))
            else:
                epsilon_Psi_as_list.append(Mod(sum(
                    epsilon_archi_product(sub_Hodge_weights[x],
                                          sub_Hodge_weights[y])
                    for y in range(len(para[0]))
                    if Mod(para[0][y][0],2)==Mod(1, 2) and para[0][y][0]<d), 2))
        epsilon_Psi_as_list.append(Mod(sum(epsilon_Psi_as_list), 2))
    epsilon_Psi = tuple(epsilon_Psi_as_list)
    if epsilon_Psi in char_dict:
        res = char_dict[epsilon_Psi]
        for x in range(len(para[0])):
            d = para[0][x][0]
            rk_x = len(para[0][x][1])
            if Mod(d,2)==Mod(1,2):
                sub_dom_weight = tuple(
                    Integer(sub_Hodge_weights[x][l]-rk_x+1/2+l)
                    for l in range(rk_x))
                res *= new_sympl_para_list[rk_x-1][sub_dom_weight]
            else:
                sub_dom_weight = tuple(Integer(sub_Hodge_weights[x][l]-rk_x+l+1)
                                  for l in range(rk_x))
                res *= new_orth_even_para_list[rk_x/2-1][sub_dom_weight]
        if para[1]>0:
            rk_x = len(para[2])
            sub_dom_weight = tuple(Integer(orth_odd_Hodge_weights[l]-rk_x+l)
                              for l in range(rk_x))
            res *= new_orth_odd_para_list[rk_x][sub_dom_weight]
        return res
    else:
        return 0

def endo_to_str_SO_even(P, sub_Hodge_weights, orth_odd_Hodge_weights):
    res = ""
    first_term = True
    for x in range(len(P[0])):
        d = P[0][x][0]
        if Mod(d,2)==0:
            if first_term:
                first_term = False
            else:
                res += " + "
            res += "S"+str(tuple(sub_Hodge_weights[x]))+"["+str(d)+"]"
        else:
            if first_term:
                first_term = False
            else:
                res += " + "
            res += "Oe"+str(tuple(sub_Hodge_weights[x]))
            if d>1:
                res += "["+str(d)+"]"
    if P[1]>0:
        if not first_term:
            res += " + "
        if len(orth_odd_Hodge_weights[0])==0:
            if P[1]==1:
                res += "1"
        else:
            res += "Oo"+str(tuple(orth_odd_Hodge_weights[0]))
        if P[1]>1:
            res += "["+str(2*P[1]-1)+"]"
        if len(orth_odd_Hodge_weights[1])==0:
            res += " + 1"
        else:
            res += " + Oo"+str(tuple(orth_odd_Hodge_weights[1]))
    return res

def endo_contrib_SO_even(n, dom_weight, para, char_dict, new_sympl_para_list,
                         new_orth_odd_para_list, new_orth_even_para_list,
                         verbose=False):
    sub_Hodge_weights = []
    epsilon_Psi_as_list = []
    for d, pos_list in para[0]:
        sub_H_wt = [dom_weight[pos]+n-1-pos-(d-1)/2 for pos in pos_list]
        sub_Hodge_weights.append(sub_H_wt)
    if para[1]>0:
        first_orth_odd_Hodge_weights = [dom_weight[pos]+n-1-pos
                                        for pos in para[2][0]]
        second_orth_odd_Hodge_weights = [dom_weight[pos]+n-1-pos-(para[1]-1)
                                         for pos in para[2][1]]
    for x in range(len(para[0])):
        d = para[0][x][0]
        # sympl[even]
        if Mod(d,2)==Mod(0,2):
            eps = Mod(sum(epsilon_archi_product(sub_Hodge_weights[x],
                                                 sub_Hodge_weights[y])
                          for y in range(len(para[0]))
                          if Mod(para[0][y][0],2)==Mod(1, 2)
                          and para[0][y][0]<d), 2)
            if para[1]>0:
                eps += epsilon_archi_product(sub_Hodge_weights[x],
                                             first_orth_odd_Hodge_weights)
                if d>2*para[1]-1:
                    eps += epsilon_archi_product(sub_Hodge_weights[x],
                                                 second_orth_odd_Hodge_weights)
                else:
                    eps += Mod(sum(w+1/2 for w in sub_Hodge_weights[x]), 2)
            epsilon_Psi_as_list.append(Mod(eps,2))
        else: # (even orth)[odd]
            epsilon_Psi_as_list.append(
                Mod(sum(epsilon_archi_product(sub_Hodge_weights[x],
                                              sub_Hodge_weights[y])
                        for y in range(len(para[0]))
                        if Mod(para[0][y][0],2)==Mod(0, 2)
                        and para[0][y][0]>d), 2))
    epsilon_Psi = tuple(epsilon_Psi_as_list)
    if epsilon_Psi in char_dict:
        res = char_dict[epsilon_Psi]
        for x in range(len(para[0])):
            d = para[0][x][0]
            rk_x = len(para[0][x][1])
            if Mod(d,2)==Mod(0,2):
                sub_dom_weight = tuple(
                    Integer(sub_Hodge_weights[x][l]-rk_x+1/2+l)
                    for l in range(rk_x))
                res *= new_sympl_para_list[rk_x-1][sub_dom_weight]
            else:
                sub_dom_weight = tuple(Integer(sub_Hodge_weights[x][l]-rk_x+l+1)
                                  for l in range(rk_x))
                res *= new_orth_even_para_list[rk_x/2-1][sub_dom_weight]
        if para[1]>0:
            rk_x = len(para[2][0])
            if rk_x>0:
                sub_dom_weight = tuple(
                    Integer(first_orth_odd_Hodge_weights[l]-rk_x+l)
                    for l in range(rk_x))
                res *= new_orth_odd_para_list[rk_x][sub_dom_weight]
            rk_x = len(para[2][1])
            if rk_x>0:
                sub_dom_weight = tuple(
                    Integer(second_orth_odd_Hodge_weights[l]-rk_x+l)
                    for l in range(rk_x))
                res *= new_orth_odd_para_list[rk_x][sub_dom_weight]
        if dom_weight[n-1]==0 and para[1]==0:
            res *= 2
        if verbose and res!=0:
            if para[1]>0:
                orth_odd_Hodge_weights = (first_orth_odd_Hodge_weights,
                                          second_orth_odd_Hodge_weights)
            else:
                orth_odd_Hodge_weights = ()
            print(endo_to_str_SO_even(para, sub_Hodge_weights,
                                      orth_odd_Hodge_weights)+": "+str(res))
        return res
    else:
        return 0

def it_dw_sub_Sp(archi_para, hodge_mw, sieg_genus, sieg_mw):
    dim = len(archi_para[0])
    indices = [0]
    for d in archi_para[0]:
        indices.append(indices[-1]+d)
    n = indices[-1]+archi_para[1]
    res = [0]*n
    if archi_para[1]==0:
        max_wt = max(hodge_mw-n, sieg_mw-1-n)
    else:
        max_wt = max(hodge_mw-n, min(sieg_genus-n, sieg_mw-1-n))
    if max_wt<0:
        return
    while True:
        yield tuple(res)
        i = dim-1
        while i>0 and res[indices[i]]==res[indices[i]-1]:
            i -= 1
        if i<0 or (i==0 and res[i]==max_wt):
            return
        for j in range(indices[i], indices[i+1]):
            res[j] += 1
        if res[0]+n<=hodge_mw:
            for j in range(indices[i+1], indices[-1]):
                res[j] = 0
        else:
            for j in range(indices[i+1], indices[-1]):
                res[j] = max(0, res[0]-sieg_genus+n)
    return

def it_dw_sub_SO_even(archi_para, hodge_mw, sieg_genus, sieg_mw):
    dim = len(archi_para[0])
    indices = [0]
    for d in archi_para[0]:
        indices.append(indices[-1]+d)
    n = indices[-1]+archi_para[1]
    res = [0]*n
    if archi_para[1]==0:
        max_wt = max(hodge_mw-n+1, sieg_mw-n)
    else:
        max_wt = hodge_mw-n+1
    if max_wt<0:
        return
    while True:
        yield tuple(res)
        i = dim-1
        while i>0 and res[indices[i]]==res[indices[i]-1]:
            i -= 1
        if i<0 or (i==0 and res[i]==max_wt):
            return
        for j in range(indices[i], indices[i+1]):
            res[j] += 1
        if res[0]+n-1<=hodge_mw:
            for j in range(indices[i+1], indices[-1]):
                res[j] = 0
        else:
            for j in range(indices[i+1], n):
                res[j] = max(1, res[0]-sieg_genus+n)
    return

def it_dw_sub_SO_odd(archi_para, hodge_mw, sieg_genus, sieg_mw):
    dim = len(archi_para[0])
    indices = [0]
    for d in archi_para[0]:
        indices.append(indices[-1]+d)
    n = indices[-1]+archi_para[1]
    res = [0]*n
    if archi_para[1]==0 and 2*n<=sieg_genus:
        max_wt = max(hodge_mw-n, sieg_mw-n-1)
    else:
        max_wt = hodge_mw-n
    if max_wt<0:
        return
    while True:
        yield tuple(res)
        i = dim-1
        while i>0 and res[indices[i]]==res[indices[i]-1]:
            i -= 1
        if i<0 or (i==0 and res[i]==max_wt):
            return
        for j in range(indices[i], indices[i+1]):
            res[j] += 1
        if res[0]+n<=hodge_mw:
            for j in range(indices[i+1], indices[-1]):
                res[j] = 0
        else:
            for j in range(indices[i+1], n):
                res[j] = max(1, res[0]-sieg_genus+n+1)
    return

def all_endo_contrib_SO_even(n, wt_bounds, extra_wts_bymask,
        new_sympl_para_list, new_orth_odd_para_list, new_orth_even_para_list,
        split=True):
    res = {}
    for global_para, archi_para, char_dict\
            in formal_global_Arthur_param_with_char_SO_even(n, split):
        m = mask_of_formal_archi_Arthur_param(archi_para)
        for dw in chain_gen(it_dw_sub_SO_even(archi_para, *wt_bounds),
                            extra_wts_bymask[m]):
            endo_contrib_para = endo_contrib_SO_even(
                n, dw, global_para, char_dict, new_sympl_para_list,
                new_orth_odd_para_list, new_orth_even_para_list)
            if not (dw in res):
                res[dw] = endo_contrib_para
            else:
                res[dw] += endo_contrib_para
    return res

def parallel_all_endo_contrib_SO_even(n, wt_bounds, extra_wts_bymask,
        new_sympl_para_list, new_orth_odd_para_list, new_orth_even_para_list,
        split=True):
    if n<=6:
        task_size = 5
        group_by = 50
    elif n==8:
        task_size = 50
        group_by = 100
    elif n==10:
        task_size = 500
        group_by = 120
    else:
        task_size = 1000
        group_by = 100
    @parallel
    def sub_par_endo_SO_even(L):
        sub_res = {}
        for global_para, archi_para, char_dict in L:
            m = mask_of_formal_archi_Arthur_param(archi_para)
            for dw in chain_gen(it_dw_sub_SO_even(archi_para, *wt_bounds),
                                extra_wts_bymask[m]):
                if Mod(sum(dw),2)!=0:
                    continue
                endo_contrib_para = endo_contrib_SO_even(
                    n, dw, global_para, char_dict, new_sympl_para_list,
                    new_orth_odd_para_list, new_orth_even_para_list)
                if not (dw in sub_res):
                    sub_res[dw] = endo_contrib_para
                else:
                    sub_res[dw] += endo_contrib_para
        return sub_res
    res = { dw : 0 for dw in chain_gen(it_dw_SO_even(n, *wt_bounds),
                                       extra_wts_bymask[0]) }
    B = iter_blocks(formal_global_Arthur_param_with_char_SO_even(n, split),
                    task_size)
    for B_list in iter_blocks(B, group_by):
        for args, sub_res in sub_par_endo_SO_even(B_list):
            for dw, m in sub_res.items():
                res[dw] += m
    return res

def all_endo_contrib_SO_odd(n, wt_bounds, extra_wts_bymask, new_sympl_para_list,
                            new_orth_odd_para_list, new_orth_even_para_list,
                            split=True):
    res = {}
    if split:
        group_type = 'B'
    else:
        group_type = 'B_cpct'
    for global_para, archi_para, char_dict\
            in formal_global_Arthur_param_with_char_BC(n, group_type):
        m = mask_of_formal_archi_Arthur_param(archi_para)
        for dw in chain_gen(it_dw_sub_SO_odd(archi_para, *wt_bounds),
                            extra_wts_bymask[m]):
            endo_contrib_para = endo_contrib_SO_odd(
                n, dw, global_para, char_dict, new_sympl_para_list,
                new_orth_odd_para_list, new_orth_even_para_list)
            if not (dw in res):
                res[dw] = endo_contrib_para
            else:
                res[dw] += endo_contrib_para
    return res

def parallel_all_endo_contrib_SO_odd(n, wt_bounds, extra_wts_bymask,
        new_sympl_para_list, new_orth_odd_para_list, new_orth_even_para_list,
        split=True):
    if split:
        group_type = 'B'
    else:
        group_type = 'B_cpct'
    if n<=6:
        task_size = 8
        group_by = 50
    elif n<=8:
        task_size = 50
        group_by = 100
    elif n<=10:
        task_size = 500
        group_by = 400
    else:
        task_size = 1000
        group_by = 100
    @parallel
    def sub_par_endo_SO_odd(L):
        sub_res = {}
        for global_para, archi_para, char_dict in L:
            m = mask_of_formal_archi_Arthur_param(archi_para)
            for dw in chain_gen(it_dw_sub_SO_odd(archi_para, *wt_bounds),
                                extra_wts_bymask[m]):
                endo_contrib_para = endo_contrib_SO_odd(
                    n, dw, global_para, char_dict, new_sympl_para_list,
                    new_orth_odd_para_list, new_orth_even_para_list)
                if not (dw in sub_res):
                    sub_res[dw] = endo_contrib_para
                else:
                    sub_res[dw] += endo_contrib_para
        return sub_res
    res = { dw : 0 for dw in chain_gen(it_dw_SO_odd(n, *wt_bounds),
                                       extra_wts_bymask[0]) }
    B = iter_blocks(formal_global_Arthur_param_with_char_BC(n, group_type),
                    task_size)
    for B_list in iter_blocks(B, group_by):
        for args, sub_res in sub_par_endo_SO_odd(B_list):
            for dw, m in sub_res.items():
                res[dw] += m
    return res

def endo_to_str_Sp(P, sub_Hodge_weights, orth_odd_Hodge_weights):
    res = ""
    for x in range(len(P[0])):
        d = P[0][x][0]
        if Mod(d,2)==Mod(0,2):
            res += "S"+str(tuple(sub_Hodge_weights[x]))+"["+str(d)+"] + "
        else:
            res += "Oe"+str(tuple(sub_Hodge_weights[x]))
            if d>1:
                res += "["+str(d)+"]"
            res += " + "
    if len(P[2])>0:
        res += "Oo"+str(tuple(orth_odd_Hodge_weights))
    if P[1]>0:
        res += "["+str(2*P[1]+1)+"]"
    elif len(P[2])==0:
        res += "1"
    return res

def endo_contrib_Sp(n, dom_weight, para, char_dict, new_sympl_para_list,
                    new_orth_odd_para_list, new_orth_even_para_list,
                    verbose=False, ret_str=False, pre_str=""):
    sub_Hodge_weights = []
    epsilon_Psi_as_list = []
    for d, pos_list in para[0]:
        sub_H_wt = [dom_weight[pos]+n-pos-(d-1)/2 for pos in pos_list]
        sub_Hodge_weights.append(sub_H_wt)
    orth_odd_Hodge_weights = [dom_weight[pos]+n-pos-para[1] for pos in para[2]]
    #print(endo_to_str_Sp(para, sub_Hodge_weights, orth_odd_Hodge_weights))
    for x in range(len(para[0])):
        d = para[0][x][0]
        if Mod(d,2)==Mod(0,2): # sympl[even]
            if d>2*para[1]+1:
                epsilon_Psi_as_list.append(
                    epsilon_archi_product(sub_Hodge_weights[x],
                                          orth_odd_Hodge_weights)\
                    +Mod(sum(w+1/2 for w in sub_Hodge_weights[x]), 2)\
                    +Mod(sum(epsilon_archi_product(sub_Hodge_weights[x],
                                                   sub_Hodge_weights[y])
                             for y in range(len(para[0]))
                             if Mod(para[0][y][0],2)==Mod(1, 2)
                             and para[0][y][0]<d), 2))
            else:
                epsilon_Psi_as_list.append(Mod(
                    sum(epsilon_archi_product(sub_Hodge_weights[x],
                                              sub_Hodge_weights[y])
                        for y in range(len(para[0]))
                        if Mod(para[0][y][0],2)==Mod(1, 2)
                        and para[0][y][0]<d), 2))
        else: # (even orth)[odd]
            epsilon_Psi_as_list.append(Mod(
                sum(epsilon_archi_product(sub_Hodge_weights[x],
                                          sub_Hodge_weights[y])
                    for y in range(len(para[0]))
                    if Mod(para[0][y][0],2)==Mod(0, 2) and para[0][y][0]>d), 2))
    epsilon_Psi = tuple(epsilon_Psi_as_list)
    if epsilon_Psi in char_dict:
        automorphic = True
        res = char_dict[epsilon_Psi]
    else:
        automorphic = False
        res = 1
    if verbose or automorphic:
        for x in range(len(para[0])):
            d = para[0][x][0]
            rk_x = len(para[0][x][1])
            if Mod(d,2)==Mod(0,2):
                sub_dom_weight = tuple(sub_Hodge_weights[x][l]-rk_x+1/2+l
                                       for l in range(rk_x))
                res *= new_sympl_para_list[rk_x-1][sub_dom_weight]
            else:
                sub_dom_weight = tuple(sub_Hodge_weights[x][l]-rk_x+l+1
                                       for l in range(rk_x))
                res *= new_orth_even_para_list[rk_x/2-1][sub_dom_weight]
        rk_x = len(para[2])
        if rk_x>0:
            sub_dom_weight = tuple(orth_odd_Hodge_weights[l]-rk_x+l
                                   for l in range(rk_x))
            res *= new_orth_odd_para_list[rk_x][sub_dom_weight]
        if ret_str and res==0:
            S = None
        if res!=0 and (verbose or ret_str):
            S = endo_to_str_Sp(para, sub_Hodge_weights, orth_odd_Hodge_weights)\
                    +": "+str(res)
            if not automorphic:
                S += " parameters but not automorphic"
            if verbose:
                print("  "+pre_str+S)
        if not automorphic:
            res = 0
        if ret_str:
            return res, S
        else:
            return res
    else:
        if ret_str:
            return 0, None
        else:
            return 0

def all_endo_contrib_Sp(n, wt_bounds, extra_wts_bymask, new_sympl_para_list,
                        new_orth_odd_para_list, new_orth_even_para_list,
                        siegel=False):
    res = {}
    if siegel:
        group_type = 'Siegel'
    else:
        group_type = 'C'
    for global_para, archi_para, char_dict\
            in formal_global_Arthur_param_with_char_BC(n, group_type):
        m = mask_of_formal_archi_Arthur_param(archi_para)
        for dw in chain_gen(it_dw_sub_Sp(archi_para, *wt_bounds),
                            extra_wts_bymask[m]):
            endo_contrib_para = endo_contrib_Sp(
                n, dw, global_para, char_dict, new_sympl_para_list,
                new_orth_odd_para_list, new_orth_even_para_list)
            if not (dw in res):
                res[dw] = endo_contrib_para
            else:
                res[dw] += endo_contrib_para
    return res

def endo_to_str_Siegel_low_3odd(P, sub_Hodge_weights, orth_odd_Hodge_weights):
    res = ""
    for x in range(len(P[0])):
        d = P[0][x][0]
        if d%2:
            res += "Oe"+str(tuple(sub_Hodge_weights[x]))
            if d>1:
                res += "["+str(d)+"]"
            res += " + "
        else:
            res += "S"+str(tuple(sub_Hodge_weights[x]))+"["+str(d)+"] + "
    a, b = P[1]
    if len(P[2][0])>0:
        res += "Oo"+str(tuple(orth_odd_Hodge_weights[0]))+" + "
    else:
        res += "1 + "
    if len(P[2][1])>0:
        res += "Oo"+str(tuple(orth_odd_Hodge_weights[1]))
    if a>1:
        res += "["+str(2*a-1)+"]"
    elif len(P[2][1])==0:
        res += "1"
    res += " + "
    if len(P[2][2])>0:
        res += "Oo"+str(tuple(orth_odd_Hodge_weights[2]))
    if b>0:
        res += "["+str(2*b+1)+"]"
    elif len(P[2][2])==0:
        res += "1"
    return res

def endo_contrib_Siegel_low_3odd(n, para, global_char, new_sympl_para_list,
        new_orth_odd_para_list, new_orth_even_para_list, verbose=False):
    # dom_weight is for SO_{2k'}. Notation: kp = k'
    kp = n-para[1][1]
    if para[3]:
        # case H1
        k = kp
        dom_weight = [0]*k
    else:
        # case H2
        k = kp+1
        dom_weight = ([1]*(2*kp-n))+([0]*(n-kp))
    sub_Hodge_weights = []
    epsilon_Psi = []
    for d, pos_list in para[0]:
        sub_H_wt = [dom_weight[pos]+kp-1-pos-(d-1)/2 for pos in pos_list]
        sub_Hodge_weights.append(sub_H_wt)
    a, b = para[1]
    first_orth_odd_Hodge_weights = [dom_weight[pos]+kp-1-pos
                                    for pos in para[2][0]]
    second_orth_odd_Hodge_weights = [dom_weight[pos]+kp-1-pos-(a-1)
                                     for pos in para[2][1]]
    third_orth_odd_Hodge_weights = [dom_weight[pos]+kp-1-pos-b
                                    for pos in para[2][2]]
    for x in range(len(para[0])):
        d = para[0][x][0]
        # sympl[even]
        if Mod(d,2)==Mod(0,2):
            eps = Mod(sum(epsilon_archi_product(sub_Hodge_weights[x],
                                                 sub_Hodge_weights[y])
                          for y in range(len(para[0]))
                          if Mod(para[0][y][0],2)==Mod(1, 2)
                          and para[0][y][0]<d), 2)
            eps += epsilon_archi_product(sub_Hodge_weights[x],
                                         first_orth_odd_Hodge_weights)
            if d>2*a-1:
                eps += epsilon_archi_product(sub_Hodge_weights[x],
                                             second_orth_odd_Hodge_weights)
            if d>2*b+1:
                eps += epsilon_archi_product(sub_Hodge_weights[x],
                                             third_orth_odd_Hodge_weights)
            if d<2*a-1 or d>2*b+1:
                eps += Mod(sum(w+1/2 for w in sub_Hodge_weights[x]), 2)
            epsilon_Psi.append(Mod(eps,2))
        else: # (even orth)[odd]
            epsilon_Psi.append(
                Mod(sum(epsilon_archi_product(sub_Hodge_weights[x],
                                              sub_Hodge_weights[y])
                        for y in range(len(para[0]))
                        if Mod(para[0][y][0],2)==Mod(0, 2)
                        and para[0][y][0]>d), 2))
    epsilon_i1 = Mod(0,2)
    for y, x in enumerate(para[0]):
        d = x[0]
        if d%2:
            continue
        epsilon_i1 += epsilon_archi_product(first_orth_odd_Hodge_weights,
                                            sub_Hodge_weights[y])\
                      +Mod(sum(w+1/2 for w in sub_Hodge_weights[y]), 2)
    epsilon_i2 = Mod(0,2)
    for y, x in enumerate(para[0]):
        d = x[0]
        if d%2 or d<2*a-1:
            continue
        epsilon_i2 += epsilon_archi_product(second_orth_odd_Hodge_weights,
                                            sub_Hodge_weights[y])\
                      +Mod(sum(w+1/2 for w in sub_Hodge_weights[y]), 2)
    epsilon_i0 = Mod(0,2)
    for y, x in enumerate(para[0]):
        d = x[0]
        if d%2 or d<2*b+1:
            continue
        epsilon_i0 += epsilon_archi_product(third_orth_odd_Hodge_weights,
                                            sub_Hodge_weights[y])\
                      +Mod(sum(w+1/2 for w in sub_Hodge_weights[y]), 2)
    epsilon_Psi += [epsilon_i1+epsilon_i2, epsilon_i0+epsilon_i2]
    if epsilon_Psi!=global_char:
        return 0
    res = 1
    for x in range(len(para[0])):
        d = para[0][x][0]
        rk_x = len(para[0][x][1])
        if Mod(d,2)==Mod(0,2):
            sub_dom_weight = tuple(
                Integer(sub_Hodge_weights[x][l]-rk_x+1/2+l)
                for l in range(rk_x))
            res *= new_sympl_para_list[rk_x-1][sub_dom_weight]
        else:
            sub_dom_weight = tuple(Integer(sub_Hodge_weights[x][l]-rk_x+l+1)
                              for l in range(rk_x))
            res *= new_orth_even_para_list[rk_x/2-1][sub_dom_weight]
    # pi_i1
    rk_x = len(para[2][0])
    if rk_x>0:
        sub_dom_weight = tuple(
            ZZ(first_orth_odd_Hodge_weights[l]-rk_x+l) for l in range(rk_x))
        res *= new_orth_odd_para_list[rk_x][sub_dom_weight]
    # pi_i2
    rk_x = len(para[2][1])
    if rk_x>0:
        sub_dom_weight = tuple(
            ZZ(second_orth_odd_Hodge_weights[l]-rk_x+l) for l in range(rk_x))
        res *= new_orth_odd_para_list[rk_x][sub_dom_weight]
    # pi_i0
    rk_x = len(para[2][2])
    if rk_x>0:
        sub_dom_weight = tuple(
            ZZ(third_orth_odd_Hodge_weights[l]-rk_x+l) for l in range(rk_x))
        res *= new_orth_odd_para_list[rk_x][sub_dom_weight]
    if verbose and res!=0:
        orth_odd_Hodge_weights = (first_orth_odd_Hodge_weights,
                                  second_orth_odd_Hodge_weights,
                                  third_orth_odd_Hodge_weights)
        print("  k="+str(k)+" "+endo_to_str_Siegel_low_3odd(para,
                sub_Hodge_weights, orth_odd_Hodge_weights)+": "+str(res))
    return res

def all_endo_contrib_Siegel_low_3odd(n, new_sympl_para_list,
        new_orth_odd_para_list, new_orth_even_para_list, max_k=100):
    res = [0 for _ in range(min(n, max_k)+1)]
    for global_para, archi_para, global_char\
            in formal_global_Arthur_param_with_char_Siegel_low_3odd(n, max_k):
        k = n-global_para[1][1]
        if not global_para[3]:
            # case H2
            k += 1
        res[k] += endo_contrib_Siegel_low_3odd(n, global_para, global_char,
                new_sympl_para_list, new_orth_odd_para_list,
                new_orth_even_para_list, verbose=True)
    return res

def all_endo_contrib_Siegel_low(n, new_sympl_para_list, new_orth_odd_para_list,
        new_orth_even_para_list, max_k=100):
    res = [0 for _ in range(min(n, max_k)+1)]
    for global_para, archi_para, char_dict\
            in formal_global_Arthur_param_with_char_BC(n, 'Siegel_low', max_k):
        a = global_para[1]
        max_pos = -1
        last_d = -1
        for d, pos_list in global_para[0]:
            pos = pos_list[-1]
            if pos>max_pos:
                max_pos = pos
                last_d = d
        # lower bound for k in cases I and [H1 with i0 in I_even], with equality
        # in the second case with a>0
        k_lower_bd = a+1+max_pos+(last_d-1)/2
        if a==0:
            if len(global_para[2])==0:
                new_lower_bd = ceil(k_lower_bd)
            else:
                new_lower_bd = max(ceil(k_lower_bd), 1+global_para[2][-1])
                # change 1+... to 2+... to exclude triple weight zero
            for k in range(new_lower_bd, len(res)):
                dw = [k-1-n for i in range(n)]
                res[k] += endo_contrib_Sp(n, dw, global_para, char_dict,
                        new_sympl_para_list, new_orth_odd_para_list,
                        new_orth_even_para_list, verbose=True, ret_str=False,
                        pre_str="k="+str(k)+" ")
        elif last_d%2 and last_d>=2*a+1:
            # case H1 with i0 in I_even and a>0
            k = ZZ(k_lower_bd)
            if k<=max_k:
                dw = [k-1-n for i in range(k-a-1)]\
                    +[k-1-n-a for i in range(n-k+1)]\
                    +[0 for i in range(a)]
                if len(dw)!=n:
                    raise ValueError
                # check that the odd pi_i has "positive Hodge weights"
                if len(global_para[2])==0 or global_para[2][-1]<k-a-1:
                    res[k] += endo_contrib_Sp(n, dw, global_para, char_dict,
                            new_sympl_para_list, new_orth_odd_para_list,
                            new_orth_even_para_list, verbose=True,
                            ret_str=False, pre_str="k="+str(k)+" ")
        if a>0:
            # case H1 with I_odd = {i0}
            k = n-a
            if k<=max_k:
                dw = [k-1-n for i in range(n)]
                #old_val = res[k]
                res[k] += endo_contrib_Sp(n, dw, global_para, char_dict,
                        new_sympl_para_list, new_orth_odd_para_list,
                        new_orth_even_para_list, verbose=True, ret_str=False,
                        pre_str="k="+str(k)+" ")
                #if old_val!=res[k]:
                #    print("H1:", dw)
            # case H2 with I_odd = {i0}
            k = n+1-a
            if k<=max_k:
                # we check the hole in Hodge weights
                is_admissible = False
                s = 0
                for d in archi_para[0]:
                    if s==2*(k-1)-n:
                        is_admissible = True
                    s += d
                if is_admissible:
                    dw = [k-1-n for i in range(2*k-n-2)]\
                        +[k-2-n for i in range(n-k+1)]
                    #old_val = res[k]
                    res[k] += endo_contrib_Sp(n, dw, global_para, char_dict,
                            new_sympl_para_list, new_orth_odd_para_list,
                            new_orth_even_para_list, verbose=True,
                            ret_str=False, pre_str="k="+str(k)+" ")
                    #if old_val!=res[k]:
                    #    print("H2:", dw)
    return res

def parallel_all_endo_contrib_Sp(n, wt_bounds, extra_wts_bymask,
        new_sympl_para_list, new_orth_odd_para_list, new_orth_even_para_list,
        siegel=False):
    if siegel:
        group_type = 'Siegel'
    else:
        group_type = 'C'
    if n<=7:
        task_size = 8
        group_by = 100
    elif n<=9:
        task_size = 50
        group_by = 300
    elif n<=10:
        task_size = 100
        group_by = 60
    else:
        task_size = 1000
        group_by = 100
    @parallel
    def sub_par_endo_Sp(L):
        sub_res = {}
        for global_para, archi_para, char_dict in L:
            m = mask_of_formal_archi_Arthur_param(archi_para)
            for dw in chain_gen(it_dw_sub_Sp(archi_para, *wt_bounds),
                                extra_wts_bymask[m]):
                if Mod(sum(dw),2)!=0:
                    continue
                endo_contrib_para = endo_contrib_Sp(
                    n, dw, global_para, char_dict, new_sympl_para_list,
                    new_orth_odd_para_list, new_orth_even_para_list)
                if not (dw in sub_res):
                    sub_res[dw] = endo_contrib_para
                else:
                    sub_res[dw] += endo_contrib_para
        return sub_res
    res = { dw : 0 for dw in chain_gen(it_dw_Sp(n, *wt_bounds),
                                       extra_wts_bymask[0]) }
    B = iter_blocks(formal_global_Arthur_param_with_char_BC(n, group_type),
                    task_size)
    for B_list in iter_blocks(B, group_by):
        for args, sub_res in sub_par_endo_Sp(B_list):
            for dw, m in sub_res.items():
                res[dw] += m
    return res

