def opposite_phi(m):
    if Mod(m,2)==1:
        return 2*m
    elif Mod(m,4)==2:
        return m/2
    else:
        return m

def opposite_cc(cc):
    res_list = [(opposite_phi(x[0]), x[1]) for x in cc]
    res_list.sort(key=lambda x: x[0])
    return tuple(res_list)

def cc_to_str(cc):
    return ' '.join([str(x[0])+" "+str(x[1]) for x in cc])

def cc_to_e_list(cc):
    R.<x> = PolynomialRing(ZZ)
    P = R(prod(R.cyclotomic_polynomial(m)^mult for m, mult in cc))
    n = P.degree()
    return [(-1)^i*P[n-i] for i in range(P.degree()+1)]

def e_from_list(dim, e_list, i):
    if i<0 or i>dim:
        return 0
    return e_list[i]

def dual_young(lam):
    i = len(lam)-1
    if i<0:
        return []
    while i>0 and lam[i]==0:
        i -= 1
    res = [i+1]*lam[i]
    ol = lam[i]
    while i>0:
        i -= 1
        while i>0 and lam[i]==lam[i+1]:
            i -= 1
        if lam[i]>ol:
            res += [i+1]*(lam[i]-ol)
            ol = lam[i]
    return res

# lam is the dual of the dominant weight
def tr_alg_rep_Sp(n, e_list, lam, l):
    return matrix(ZZ, l, l, lambda i,j:
                  e_from_list(2*n, e_list, lam[i]-i+j)\
                  -e_from_list(2*n, e_list, lam[i]-i-2-j)).det()

# NB: for SO_even and a dominant weight that is not invariant under outer
# automorphism, this computes the sum over the orbit (which has 2 elements)
# under the outer aut group. So for SO_even we are really computing the trace in
# an irreducible representation of O
def tr_alg_rep_SO(dim, e_list, lam, l):
    if l==0:
        return 1
    return matrix(ZZ, l, l, lambda i,j:
                  e_from_list(dim, e_list, lam[i]-i+j)\
                  +e_from_list(dim, e_list, lam[i]-i-j)).det()/2

def contrib_ell(group_type, n, wt_bounds, extra_wts, cc_mass_dict):
    if group_type=='B':
        el_mass_list = [(cc_to_e_list(cc), mass)
                        for cc, mass in cc_mass_dict.items()]
    else:
        # avoid redundant computations by grouping orbits under the center
        already_opp_set = set()
        el_mass_list = []
        for cc, mass in cc_mass_dict.items():
            opp_cc = opposite_cc(cc)
            if group_type=='D' and cc[0][0]>2: 
                # two cc's have this characteristic polynomial
                mu = 2
            else:
                mu = 1
            if opp_cc!=cc:
                if not (cc in already_opp_set):
                    el_mass_list.append((cc_to_e_list(cc), 2*mu*mass))
                    already_opp_set.add(opp_cc)
            else:
                el_mass_list.append((cc_to_e_list(cc), mu*mass))
    num_cc = len(el_mass_list)
    # these choices are rather arbitrary
    if num_cc>200:
        task_size = (num_cc/8).floor()
    elif num_cc>50:
        task_size = (num_cc/4).floor()
    else:
        task_size = num_cc
    el_mass_sublists_list = list(iter_blocks(el_mass_list, task_size))
    if group_type=='B':
        it_dw = it_dw_SO_odd
        tr_fun = lambda el, lam, dw0: tr_alg_rep_SO(2*n+1, el, lam, dw0)
    elif group_type=='C':
        it_dw = it_dw_Sp
        tr_fun = lambda el, lam, dw0: tr_alg_rep_Sp(n, el, lam, dw0)
    elif group_type=='D':
        it_dw = it_dw_SO_even
        tr_fun = lambda el, lam, dw0: tr_alg_rep_SO(2*n, el, lam, dw0)
    @parallel(ncpus=4)
    def helper(el_mass_sublist):
        sub_res = {}
        for dw in chain_gen(it_dw(n, *wt_bounds), extra_wts):
            if group_type!='B' and Mod(sum(dw),2)!=0:
                continue
            lam = dual_young(dw)
            if group_type=='D' and dw[n-1]>0:
                sub_res[dw] = sum(mass*tr_fun(el, lam, dw[0])
                                  for el, mass in el_mass_sublist)/2
            else:
                sub_res[dw] = sum(mass*tr_fun(el, lam, dw[0])
                                  for el, mass in el_mass_sublist)
        return sub_res
    res = {dw : 0 for dw in chain_gen(it_dw(n, *wt_bounds), extra_wts)}
    for el_mass_sublist, sub_res in helper(el_mass_sublists_list):
        for dw, val in sub_res.items():
            res[dw] += val
    return res

def contrib_ell_SO_odd(n, wt_bounds, extra_wts, cc_mass_dict):
    return contrib_ell('B', n, wt_bounds, extra_wts, cc_mass_dict)

def contrib_ell_Sp(n, wt_bounds, extra_wts, cc_mass_dict):
    return contrib_ell('C', n, wt_bounds, extra_wts, cc_mass_dict)

def contrib_ell_SO_even(n, wt_bounds, extra_wts, cc_mass_dict):
    return contrib_ell('D', n, wt_bounds, extra_wts, cc_mass_dict)

def mat_tr_alg_rep(group_type, n, dw_list, cc_list):
    if group_type=='B':
        cc_mass_list = [(cc, 1) for cc in cc_list]
    else:
        # avoid redundant computations by grouping orbits under the center
        already_opp_set = set()
        cc_mass_list = []
        for cc in cc_list:
            opp_cc = opposite_cc(cc)
            if group_type=='D' and cc[0][0]>2: 
                # two cc's have this characteristic polynomial
                mu = 2
            else:
                mu = 1
            if opp_cc!=cc:
                if not (cc in already_opp_set):
                    cc_mass_list.append((cc, 2*mu))
                    already_opp_set.add(opp_cc)
            else:
                cc_mass_list.append((cc, mu))
    num_cc = len(cc_mass_list)
    # these choices are rather arbitrary
    if num_cc>200:
        task_size = (num_cc/8).floor()
    elif num_cc>50:
        task_size = (num_cc/4).floor()
    else:
        task_size = num_cc
    cc_mass_sublists_list = list(iter_blocks(cc_mass_list, task_size))
    if group_type=='B':
        tr_fun = lambda el, lam, dw0: tr_alg_rep_SO(2*n+1, el, lam, dw0)
    elif group_type=='C':
        tr_fun = lambda el, lam, dw0: tr_alg_rep_Sp(n, el, lam, dw0)
    elif group_type=='D':
        tr_fun = lambda el, lam, dw0: tr_alg_rep_SO(2*n, el, lam, dw0)
    num_dw = len(dw_list)
    @parallel(ncpus=4)
    def helper(cc_mass_sublist):
        el_mass_sublist = [(cc_to_e_list(cc), mass)
                           for cc, mass in cc_mass_sublist]
        sub_M = matrix(ZZ, len(cc_mass_sublist), num_dw)
        for j, dw in enumerate(dw_list):
            if group_type!='B' and Mod(sum(dw),2)!=0:
                raise ValueError("useless weight")
            lam = dual_young(dw)
            if group_type=='D' and dw[n-1]>0:
                pond = 1/2
            else:
                pond = 1
            for i, (el, mass) in enumerate(el_mass_sublist):
                sub_M[i,j] = mass*tr_fun(el, lam, dw[0])*pond
        return sub_M
    M = matrix(ZZ, 0, num_dw)
    new_cc_list = []
    for ((cc_mass_sublist,), _), sub_M in helper(cc_mass_sublists_list):
        new_cc_list += [cc for cc, _ in cc_mass_sublist]
        M = block_matrix([[M], [sub_M]], subdivide=False)
    return new_cc_list, M

# compute matrix of traces for conjugacy classes in cc_list and dominant weights
# in dw_list
def mat_tr_alg_rep_SO_odd(n, dw_list, cc_list):
    M = matrix(ZZ, len(cc_list), len(dw_list))
    j = 0
    for j, dw in enumerate(dw_list):
        lam = dual_young(dw)
        for i, el in enumerate(cc_list):
            M[i,j] = tr_alg_rep_SO(2*n+1, cc_to_e_list(el), lam, dw[0])
    # we also return the argument cc_list for consistency with other cases (see
    # below)
    return (cc_list, M)

# for Sp and SO_even we take the center into account
def mat_tr_alg_rep_CD(n, wts, cc_list, tr_fn):
    # will contain representatives for action of center on cc_list
    cc_list_modZ = []
    already_opp_set = set()
    # will contain char polys of elements of cc_list_modZ, plus weight according
    # to orbit size
    el_mass_list = []
    for cc in cc_list:
        opp_cc = opposite_cc(cc)
        if opp_cc!=cc:
            if not (cc in already_opp_set):
                el_mass_list.append((cc_to_e_list(cc), 2))
                cc_list_modZ.append(cc)
                already_opp_set.add(opp_cc)
        else:
            el_mass_list.append((cc_to_e_list(cc), 1))
            cc_list_modZ.append(cc)
    num_cols = sum(1 for dw in wts if Mod(sum(dw),2)==0)
    M = matrix(ZZ, len(cc_list_modZ), num_cols)
    j = 0
    for dw in wts:
        if Mod(sum(dw),2)==0:
            lam = dual_young(dw)
            for i, (el, mass) in enumerate(el_mass_list):
                M[i,j] = mass*tr_fn(n, el, lam, dw[0])
            j += 1
    return (cc_list_modZ, M)

def mat_tr_alg_rep_SO_even(r, wts, cc_list):
    tr_fn = lambda n, el, lam, k1 : tr_alg_rep_SO(2*n, el, lam, k1)
    return mat_tr_alg_rep_CD(r, wts, cc_list, tr_fn)

def mat_tr_alg_rep_Sp(r, wts, cc_list):
    return mat_tr_alg_rep_CD(r, wts, cc_list, tr_alg_rep_Sp)
